#include "scanDefines.hlsl"

cbuffer ScanParameters
{
  uint baseLevel;
  uint topLevel;
  uint offsetA; // offset to current layer
  uint offsetB; // offset to layer above
  uint offsetC; // where to write the extra element in output
  uint gridWidth;
  uint N;
  uint B;
  uint inputWidth;
};

RWStructuredBuffer<uint> output;
RWStructuredBuffer<uint> scratch;
#ifdef SCAN_VERIFY
RWStructuredBuffer<uint> debugInput;
RWStructuredBuffer<uint> debugOutput;
#endif
Texture2D<float> inputTexture;
groupshared uint shmem[SCAN_BLOCKSIZE];

[numthreads(SCAN_WARPSIZE, 4, 1)]
void main(uint3 threadIdx : SV_GroupThreadID,
          uint3 blockIdx : SV_GroupID)
{
  const uint wt = threadIdx.x;
  const uint warp = threadIdx.y;
  const uint t = warp*SCAN_WARPSIZE + wt;
  const uint b = blockIdx.y*gridWidth + blockIdx.x;
  if (B <= b) {
    return;
  }

  const uint4 ix = (uint4)(4 * (SCAN_BLOCKSIZE*b + t)) + uint4(0, 1, 2, 3);

  // Start by summing 4 input values to shmem[thread].
  uint4 values;
  if (baseLevel) {
    if (ix.x < N) {
      uint4 i = ix % (uint4)inputWidth;
      uint4 j = ix / (uint4)inputWidth;
      if (ix.w < N) {
        values.x = inputTexture.Load(int3(i.x, j.x, 0));
        values.y = inputTexture.Load(int3(i.y, j.y, 0));
        values.z = inputTexture.Load(int3(i.z, j.z, 0));
        values.w = inputTexture.Load(int3(i.w, j.w, 0));
      }
      else {
        values.x = inputTexture.Load(int3(i.x, j.x, 0));
        values.y = ix.y < N ? inputTexture.Load(int3(i.y, j.y, 0)) : 0;
        values.z = ix.z < N ? inputTexture.Load(int3(i.z, j.z, 0)) : 0;
        values.w = 0;
      }

#ifdef SCAN_VERIFY
      if (ix.x < N) debugInput[ix.x] = inputTexture.Load(int3(i.x, j.x, 0));
      if (ix.y < N) debugInput[ix.y] = inputTexture.Load(int3(i.y, j.y, 0));
      if (ix.z < N) debugInput[ix.z] = inputTexture.Load(int3(i.z, j.z, 0));
      if (ix.w < N) debugInput[ix.w] = inputTexture.Load(int3(i.w, j.w, 0));
#endif
    }
    else {
      values = (uint4)0;
    }
  }
  else {
    if (ix.x < N) {
      const uint4 jx = (uint4)offsetA + ix;
      if (ix.w < N) {
        values.x = scratch[jx.x];
        values.y = scratch[jx.y];
        values.z = scratch[jx.z];
        values.w = scratch[jx.w];
      }
      else {
        values.x = scratch[jx.x];
        values.y = ix.y < N ? scratch[jx.y] : 0;
        values.z = ix.z < N ? scratch[jx.z] : 0;
        values.w = 0;
      }
    }
    else {
      values = (uint4)0;
    }

  }
  shmem[t] = values.x + values.y + values.z + values.w;

  // Each warp builds its own inclusive scan:
  //          [ 0 |   1 |     2 |       3 |         4 |           5 |           6 |          7]
  // l=0; o=1 [   | 0+1 |   1+2 |     2+3 |       3+4 |         4+5 |         5+6 |        6+7]
  // l=1; o=2 [   |     | 0+1+2 | 0+1+2+3 |   1+2+3+4 |     2+3+4+5 |     3+4+5+6 |    4+5+6+7]
  // l=2; o=4 [   |     |       |         | 0+1...3+4 | 0+1+...+4+5 | 0+1+...+5+6 | 0+1+...6+7]
#ifdef SCAN_WARP_LOCKSTEP
  GroupMemoryBarrier();
  [unroll]
  for (uint l = 0; l < SCAN_WARPSIZE_L2; l++) {
    const uint o = 1 << l;
    if (o <= wt) {
      uint a = shmem[t - o] + shmem[t];
      GroupMemoryBarrier();
      shmem[t] = a;
      GroupMemoryBarrier();
    }
  }
  GroupMemoryBarrierWithGroupSync();

#else
  GroupMemoryBarrierWithGroupSync();
  [unroll]
  for (uint l = 0; l < SCAN_WARPSIZE_L2; l++) {
    const uint o = 1 << l;
    uint a = 0;
    if (o <= wt) {
      a = shmem[t - o] + shmem[t];
    }
    GroupMemoryBarrierWithGroupSync();
    if (o <= wt) {
      shmem[t] = a;
    }
    GroupMemoryBarrierWithGroupSync();
  }
#endif
  // Thread 0 pulls per-chunk scan results from level above
  // and distributes cumulative sums 
  if (t == 0) {
    uint s = topLevel ? 0 : scratch[offsetB + b]; // Per chunks scan from level above
    uint n = shmem[1 * SCAN_WARPSIZE - 1];             // Local element sum for warp 0
    shmem[1 * SCAN_WARPSIZE - 1] = s;                  // Set offset for warp 0 in last element.
    s = s + n;                                    // One step of exclusive scan
    n = shmem[2 * SCAN_WARPSIZE - 1];                  // Local element sum for warp 1
    shmem[2 * SCAN_WARPSIZE - 1] = s;
    s = s + n;
    n = shmem[3 * SCAN_WARPSIZE - 1];
    shmem[3 * SCAN_WARPSIZE - 1] = s;
    s = s + n;
    n = shmem[4 * SCAN_WARPSIZE - 1];
    shmem[4 * SCAN_WARPSIZE - 1] = s;
    if (topLevel) {
      s = s + n;
      output[offsetC] = s;
#ifdef SCAN_VERIFY
      debugOutput[offsetC] = s;
#endif
    }
  }

  GroupMemoryBarrierWithGroupSync();

  // All threads fetch the offset for its warp and update its do a mini-scan for
  // the four values it is responsible for.
  {
    const uint warpOffset = shmem[(warp + 1)*SCAN_WARPSIZE - 1];

    uint a = (wt == 0 ? 0 : shmem[t - 1]) + warpOffset;
    uint b = a + values.x;
    uint c = b + values.y;
    uint d = c + values.z;

    if (baseLevel) {
      if (ix.w < N) {
        output[ix.x] = a;
        output[ix.y] = b;
        output[ix.z] = c;
        output[ix.w] = d;

#ifdef SCAN_VERIFY
        debugOutput[ix.x] = a;
        debugOutput[ix.y] = b;
        debugOutput[ix.z] = c;
        debugOutput[ix.w] = d;
#endif
      }
      else if (ix.x < N) {
        output[ix.x] = a;
        if (ix.y < N) output[ix.y] = b;
        if (ix.z < N) output[ix.z] = c;

#ifdef SCAN_VERIFY
        debugOutput[ix.x] = a;
        if (ix.y < N) debugOutput[ix.y] = b;
        if (ix.z < N) debugOutput[ix.z] = c;
#endif
      }
    }
    else {
      const uint4 jx = (uint4)offsetA + ix;
      if (ix.w < N) {
        scratch[jx.x] = a;
        scratch[jx.y] = b;
        scratch[jx.z] = c;
        scratch[jx.w] = d;
      }
      else if (ix.x < N) {
        scratch[jx.x] = a;
        if (ix.y < N) scratch[jx.y] = b;
        if (ix.z < N) scratch[jx.z] = c;
      }
    }
  }
}
