#include "Engine/TransparencyBlendCommon.hlsl"
#include "Engine/PhongLighting.hlsl"

Texture2D depthTexture;
SamplerState linearSampler;

ByteAddressBuffer               List;
StructuredBuffer<FragmentNode>  Node;
StructuredBuffer<FragmentData>  Data;

PixelOut error()
{
  PixelOut Out;
  Out.Color.rgb = float3(1, 0, 0);
  Out.Color.a = 1.0;
  return Out;
}

float3 euclidean(float4 h)
{
  return (1.0 / h.w)*h.xyz;
}

PixelOut main(VertexIn In, float4 ScreenPos : SV_Position)
{
  const float backgroundDepth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  uint startOffset = 4 * ((uint)ScreenPos.y/HeightDivisor * Width + (uint)ScreenPos.x/WidthDivisor);
  uint rx = List.Load(startOffset);
  uint ix = rx & 0x7fffffffu;

  if (NodeAlloc <= ix) {
    PixelOut BackOut = { float4(0, 0, 0, 1) };
    return BackOut;
  }

  // Plow through the full fragment list for a pixel location _once_, while maintaining
  // a depth-sorted array of the MAX_SAMPLES nearest entries.
  uint indices[MAX_SAMPLES];
  float depths[MAX_SAMPLES];
  int tailIx = 0;
  indices[0] = rx;
  depths[0] = Node[ix].Depth;
  rx = Node[ix].Next;
  ix = rx & 0x7fffffffu;

  uint maxIter = 100;
  for(uint currentIter = 0; currentIter != maxIter && ix < NodeAlloc; currentIter++) {
    float dp = Node[ix].Depth;
    bool notFull = tailIx < MAX_SAMPLES - 1;
#ifdef COGS_REVERSE_DEPTH
    bool closerThanTail = dp > depths[tailIx];
#else
    bool closerThanTail = dp < depths[tailIx];
#endif
    if (notFull || closerThanTail) {
      tailIx = tailIx + (notFull ? 1 : 0);
      int j = tailIx;
      for (int i = tailIx - 1; i >= 0; i--) {
#ifdef COGS_REVERSE_DEPTH
        if (dp > depths[i]) {
#else
        if (dp < depths[i]) {
#endif
          indices[i + 1] = indices[i];
          depths[i + 1] = depths[i];
          j = i;
        }
      }
      indices[j] = rx;
      depths[j] = dp;
    }
    rx = Node[ix].Next;
    ix = rx & 0x7fffffffu;
  }

  int stackPtr = 0;
  float4 result = float4(0, 0, 0, 1);
  float4 t = mul(float4(In.ScreenPosition.xy, depths[0], 1), inverseProjectionMatrix);
  float3 viewSpacePosA = euclidean(t);
  float3 worldSpacePosA = euclidean(mul(t, inverseViewMatrix));

  // Only take extra step to opaque if we haven't discarded any samples.
  int M = min(tailIx + 1, MAX_SAMPLES - 1);
  for (int i = 0; i < M && (result.a > (0.5/256.0)); i++) {
    rx = indices[i];
    ix = rx & 0x7fffffffu;
    FragmentData c = Data[ix];

    float depth = i < tailIx ? depths[i + 1] : backgroundDepth;
    float4 t = mul(float4(In.ScreenPosition.xy, depth, 1), inverseProjectionMatrix);
    float3 viewSpacePosB = euclidean(t);
    float3 worldSpacePosB = euclidean(mul(t, inverseViewMatrix));

    VolumeLighting volumeLighting = (VolumeLighting)0;
    int LightModel = c.Emissive >> 24;
    bool closed = (LightModel & 16) != 0;
    if (closed) {
      bool entering = (rx & 0x80000000u) != 0;

      if (entering) {
        indices[stackPtr] = ix; // note, store masked index in stack.
        stackPtr++;
      }
      else {
        stackPtr--;
        if (stackPtr < 1) {
          // If stackPtr == 0, we're skipping empty space. On the other hand,
          // if stackPtr < 0, we have exited more volumes than we have entered,
          // and that means that we're either inside volmes (entering surfaces
          // has been clipped), or we have messed up the orientation of the
          // surfaces.
          //if (stackPtr < 0) {
          //  return error();
          //}
          viewSpacePosA = viewSpacePosB;
          worldSpacePosA = worldSpacePosB;
          continue;
        }
        ix = indices[stackPtr - 1];
        c = Data[ix];
      }

      // For solid shapes, we do the lighting at the entering site
      volumeLighting.ViewSpacePosition = viewSpacePosA;
    }
    else {
      // For open shapes, the current site is used.
      volumeLighting.ViewSpacePosition = viewSpacePosB;
    }

    unpack(volumeLighting.Normal,
           volumeLighting.LightModel,
           volumeLighting.Color.a,
           volumeLighting.Emissive,
           volumeLighting.Color.rgb,
           volumeLighting.Specular,
           volumeLighting.SpecularPower,
           c);

    volumeLighting.WorldPositionA = float4(worldSpacePosA, 1.0);
    volumeLighting.WorldPositionB = float4(worldSpacePosB, 1.0);

    float4 shaded = float4(DoLighting(volumeLighting, ScreenPos).rgb, 1);

    // 1 = completely transparent
    // 0 = completely opaque
    float transparency;
    if (closed) {
      float l = distance(worldSpacePosA, worldSpacePosB);
      transparency = exp(-beersScaleFactor*volumeLighting.Color.a*l);
    }
    else {
      transparency = 1 - volumeLighting.Color.a;
    }
    viewSpacePosA = viewSpacePosB;
    worldSpacePosA = worldSpacePosB;

    float3 contribution = shaded.rgb - transparency*shaded.rgb;
    result.rgb = result.rgb + result.a*contribution;
    result.a = result.a*transparency;
  }

  PixelOut Out = { result };
  return Out;
}
