#include "Engine/Common.hlsl"
#include "Engine/PhongLighting.hlsl"
#include "Noise.hlsl"
#include "Bayer.hlsl"
#include "Random.hlsl"

Texture2D colorTexture;
Texture2D emissiveTexture;
Texture2D specularTexture;
Texture2D normalTexture;
Texture2D depthTexture;

TextureCube superSeaRadiance;

SamplerState superSeaSampler;

Texture2D oceanReal;
SamplerState oceanRealSampler;

Texture2D oceanImag;
SamplerState oceanImagSampler;

Texture2D oceanTangent;
SamplerState oceanTangentSampler;

float4x4 inverseViewProjectionMatrix;

cbuffer OceanParameters
{
  float fftTileScale;
  float significantWaveHeight;
};

struct VertexIn
{
  float4 Position : SV_Position;
  float4 ScreenPosition : TEXCOORD0;
  float2 TexCoords : TEXCOORD1;
};

float4 oceanParameters(float2 p)
{
  float tilePeriod = 30;
  float phaseShiftNoiseFrequency = 3.0;
  float phaseShiftNoisePeriod = phaseShiftNoiseFrequency*tilePeriod;

  p *= fftTileScale;

  float theta = 2.1 * pnoise(phaseShiftNoiseFrequency * p, (float2)(phaseShiftNoisePeriod));
  float2 phase = float2(cos(theta), sin(theta));

  return float4(phase, p);
}

float3 waveDisplace(float2 p)
{
  float4 w = oceanParameters(p);

  float3 real = oceanReal.SampleLevel(oceanRealSampler, w.zw, 0).rgb;
  float3 imag = oceanImag.SampleLevel(oceanImagSampler, w.zw, 0).rgb;

  float z = -(real.z * w.x - imag.z * w.y);
  float2 xy = real.xy * w.y + imag.xy * w.x;

  return significantWaveHeight * float3(xy, z);
}

float3 oceanNormal(float2 p)
{
  float4 w = oceanParameters(p);

  float4 tmp = significantWaveHeight * (oceanTangent.SampleLevel(oceanTangentSampler, w.zw, 0.0)
                                        + 0.25 * oceanTangent.SampleLevel(oceanTangentSampler, 16.0 * w.zw, 0.0));

  float2 rxy = (tmp.xz * w.x - tmp.yw * w.y); // Not completely correct, but seems to increase sharpness a bit
  float2 xy = tmp.xz * w.y + tmp.yw * w.x;

  float3 du = float3(1.0 + rxy.x, 0.0, xy.x);
  float3 dv = float3(0.0, 1.0 + rxy.y, xy.y);
  return normalize(cross(du, dv));
}


float4 main(VertexIn In, float4 ScreenPos : SV_Position) : SV_Target
{
  const float3 sun = normalize(float3(1, -1, 1));
  const int angularSamples = 16;
  const int radialSamples = 32;

  const float overWaterInfluence = 5.0;
  const float overWaterSampleScale = 0.9;
  const float underWaterInfluence = 15.0;
  const float underWaterBlendKnee = 0.25;
  const float underWaterSampleScale = 1.0;
  const float3 underWaterColor = float3(0.001, 0.01, 0.02);
  const float3 reflectedColor = float3(1, 1, 0.9);
  const float3 causticColor = float3(0.134, 0.213, 0.227f);

  const float4 color = colorTexture.Load(int3(ScreenPos.xy, 0));
  const float4 emissive = emissiveTexture.Load(int3(ScreenPos.xy, 0));
  const float4 specular = specularTexture.Load(int3(ScreenPos.xy, 0));
  const float3 normal = normalTexture.Load(int3(ScreenPos.xy, 0)).xyz;
  const float depth = depthTexture.Load(int3(ScreenPos.xy, 0)).r;

  float4 worldPosition = mul(float4(In.ScreenPosition.xy, depth, 1), inverseViewProjectionMatrix);
  worldPosition.xyz *= (1.0 / worldPosition.w);

  // Early exit, unshaded pixel or above overwater reflections.
  int LightModel = (int)(255 * specular.w);

  if (((LightModel & COGS_LIGHT_MASK_INDIRECT) == 0) || (overWaterInfluence + significantWaveHeight < worldPosition.z)) {
    return float4(0, 0, 0, 1);
  }

  // Early exit, below underwater caustics, only attenuation.
  else if (worldPosition.z < -underWaterInfluence - significantWaveHeight) {
    float depth = -worldPosition.z;
    float blend = underWaterBlendKnee;
    return float4(0, 0, 0, 1.0 / (1.0 + 0.01*depth*depth));
    //return float4((1 - blend)*(1 - blend)*(1 - blend)*underWaterColor*color.rgb, 1.0 / (1.0 + 0.01*depth*depth));
  }

  float waveHeight = waveDisplace(worldPosition.xy).z;
  uint state = wangHash((uint)ScreenPos.x + (((uint)ScreenPos.y)<<16));

  float3 n = normal;

  // Underwater caustics
  if (worldPosition.z <= waveHeight) {
    float depth = waveHeight - worldPosition.z + rand(state);
    float blend_ = max(0, 1.0 - (1.0 / underWaterInfluence)*depth);
    float blend = max(underWaterBlendKnee, 1.0 - (1.0 / underWaterInfluence)*depth);
    float4 rv = float4((1 - blend)*(1 - blend)*(1 - blend)*underWaterColor*color.rgb, 1.0 / (1.0 + 0.01*worldPosition.z*worldPosition.z));

#ifdef SIMPLE
    if (false) {
#else
    if ((LightModel & COGS_LIGHT_MASK_INDIRECT) == COGS_LIGHT_MASK_INDIRECT) {
#endif
      int angularSamples_ = 8;// max(4, (int)angularSamples*blend);

      float w = 0;
      float3 acc = float3(0, 0, 0);
      for (int j = 0; j < angularSamples_; j++) {
        float2 v;
        sincos((6.28318530718 / angularSamples_)*(j /*+ 1.0*rand(state)*/), v.y, v.x);

        if (dot(n.xy, v) < 0.0) continue;

        int radialSamples_ = max(2, (int)radialSamples*blend);

        for (int i = 0; i < radialSamples_; i++) {

          float3 c = float3(worldPosition.xy + depth*(underWaterSampleScale / (radialSamples_))*(i + rand(state))*v, waveHeight);

          float3 q = c - worldPosition.xyz;
          float3 r = normalize(q);
          float r2 = length(q);

          float3 l = refract(r, -oceanNormal(c.xy), 1.3);

          if (0 < dot(l, l)) {
            float3 radiance = 0.2 * superSeaRadiance.SampleLevel(environmentRadianceSampler, l.xzy, 0.0).rgb;

            float sun_cosA = 10000 * max(0, dot(l, sun) - 0.999);
            float sun_cosB = 100000 * max(0, dot(l, sun) - 0.9999);
            float sun_cosC = 500000 * max(0, dot(l, sun) - 0.99998);

            acc += max(0, dot(n, r)) * min(1, (1.0 / (1 + r2))) * (radiance + 5 * sun_cosA + 10 * sun_cosB + 50 * sun_cosC).rgb;
          }
        }
      }

      float3 i = max(0, 1 - abs(n.z))* blend_*blend_ * 500 * blend_ * (1.0 / (32 * 32))*acc;
      rv.rgb += i * causticColor;
    }
    else {
      float acc = 100 * pow(max(0, dot(n, sun)), 5.0);
      float i = max(0, 1 - abs(n.z))*blend_* blend_*blend_ * 500 * blend_ * (1.0 / (32 * 32))*acc;
      rv.rgb += i*causticColor;
    }

    return rv;
  }

  // Overwater indirect reflections
  else {
    return float4(0, 0, 0, 1);
/*
    float blend = max(0, 1.0 - (1.0 / overWaterInfluence)*(worldPosition.z - waveHeight));
    float height = worldPosition.z - waveHeight + rand(state);

    float acc = 0;
    for (int j = 0; j < angularSamples; j++) {

      float2 v;
      sincos((6.28318530718 / angularSamples)*(j + rand(state)), v.y, v.x);

      if (dot(n.xy, v) < 0.0) continue;

      for (int i = 0; i < radialSamples; i++) {

        float3 c = float3(worldPosition.xy + height*(overWaterSampleScale / radialSamples)*(i + rand(state))*v, waveHeight);

        float3 l = reflect(normalize(c - worldPosition.xyz), oceanNormal(c.xy));

        float3 radiance = max((float3)0, superSeaRadiance.SampleLevel(superSeaSampler, l.xzy, 0.0).rgb - (float3)-0.03);
        acc += 0.5*radiance + 0.5*pow(max(0.0, dot(l, sun)), 500.0);

      }
    }

    float i = max(0, 1 - abs(n.z))* blend*blend * 100 * (1.0 / (32 * 32))*acc;

    return float4(i*reflectedColor, 1);*/
  }

}
